# -*- coding: utf-8 -*-
"""
Subtractor: A Connector class that 'immediately' removes population.

The purpose of this connector is to include the effect of removal from
populations. For example it is used to track those currently in hospital, instead
of just monitoring the total hospitalized.

A scale factor can be applied. The scale factor should be positive.

@author: karlen
"""
from scipy import stats

from pypmca.Connector import Connector
from pypmca.Population import Population
from pypmca.Parameter import Parameter

class Subtractor(Connector):
    """
    Subtractor removes from the from_population, the newcomers to the
    to_population:
        - connector_name: string, short descriptor
        - from_population: Population object which will be reduced
        by the newcomers to the to_population
        - to_population: Population object having received the newcomers
        of the from_population totals. Its future is not affected by
        this connector.
        - scale_factor: Parameter object that multiplies expectation. Data treats fraction as binomial
        - ratio_populations: List of 2 populations, the ratio of those populations is applied as
        a scale factor
    """

    def __init__(self, connector_name: str, from_population: Population, to_population: Population,
                 scale_factor: Parameter = None, ratio_populations: list = None):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, Population):
            raise TypeError('Subtractor (' + self.name +
                            ') from_population must be a Population object')

        if not isinstance(to_population, Population):
            raise TypeError('Subtractor (' + self.name +
                            ') to_population must be a Population object')

        if hasattr(from_population, 'report_noise') and from_population.report_noise:
            raise TypeError('Subtractor (' + self.name +
                            ') from_population cannot have report_noise set to True')

        if scale_factor is not None:
            if not isinstance(scale_factor, Parameter):
                raise TypeError('Subtractor('+self.name+
                            ') scale_factor must be a Parameter object')
        self.scale_factor = scale_factor

        if ratio_populations is not None:
            if not isinstance(ratio_populations, list) or len(ratio_populations) != 2:
                raise TypeError('Subtractor('+self.name+
                            ') ratio_populations must be a list of two Population objects')
            for i in range(2):
                if not isinstance(ratio_populations[i], Population):
                    raise TypeError('Subtractor(' + self.name +
                                    ') ratio_populations must be a list of two Population objects')
        self.ratio_populations = ratio_populations

        # Identify the from_population as one not to derive daily numbers
        self.from_population.monotonic = False

    def update_expectation(self):
        """
        do subtraction for expectation
        """
        reduction = 0
        if len(self.to_population.future) > 0:
            reduction = -1 * self.to_population.future[0]

        if reduction < 0:
            if getattr(self, "scale_factor", None) is not None:
                reduction = self.scale_factor.get_value() * reduction
            if getattr(self, "ratio_populations", None) is not None:
                if self.ratio_populations[1].history[-1] > 0.:
                    ratio = self.ratio_populations[0].history[-1] / self.ratio_populations[1].history[-1]
                    reduction = ratio * reduction

            self.from_population.update_future_fast(reduction)

    def update_data(self):
        """
        do subtraction
        """
        reduction = 0
        if len(self.to_population.future) > 0:
            reduction = -1 * self.to_population.future[0]

        if reduction < 0:
            scale = 1
            if getattr(self, "scale_factor", None) is not None:
                scale = self.scale_factor.get_value() * scale
            if getattr(self, "ratio_populations", None) is not None:
                if self.ratio_populations[1].history[-1] > 0.:
                    ratio = 1. * self.ratio_populations[0].history[-1] / self.ratio_populations[1].history[-1]
                    scale = ratio * scale

            if scale != 1:
                iscale = int(scale)
                i_reduction = iscale * reduction
                fscale = scale - iscale
                f_reduction = stats.binom.rvs(reduction, fscale)
                reduction = i_reduction + f_reduction

            self.from_population.update_future_fast(reduction)



# -*- coding: utf-8 -*-
"""
Adder: A Connector class that 'immediately' adds population.

The purpose of this connector is to copy incoming population into the
from_population to the to_population.

A scale factor can be applied. The scale factor should be positive.

@author: karlen
"""
from scipy import stats

from pypmca.Connector import Connector
from pypmca.Population import Population
from pypmca.Parameter import Parameter

class Adder(Connector):
    """
    Adder copies the new incoming population into the from_population, 
    to the to_population:
        - connector_name: string, short descriptor
        - from_population: Population object which has newcomers. Its future
        is not affected by this connector.
        - to_population: Population object that receives the newcomers
        entering from_population.
        - scale_factor: Parameter object that multiplies expectation. Data treats fraction as binomial
        - ratio_populations: List of 2 populations, the ratio of those populations is applied as
        a scale factor
    """

    def __init__(self, connector_name: str, from_population: Population, to_population: Population,
                 scale_factor: Parameter = None, ratio_populations: list = None):
        """Constructor
        """
        super().__init__(connector_name, from_population, to_population)

        if not isinstance(from_population, Population):
            raise TypeError('Adder('+self.name+
                            ') from_population must be a Population object')

        if not isinstance(to_population, Population):
            raise TypeError('Adder ('+self.name+
                            ') to_population must be a Population object')

        if scale_factor is not None:
            if not isinstance(scale_factor, Parameter):
                raise TypeError('Adder('+self.name+
                            ') scale_factor must be a Parameter object')
        self.scale_factor = scale_factor

        if ratio_populations is not None:
            if not isinstance(ratio_populations, list) or len(ratio_populations) != 2:
                raise TypeError('Adder('+self.name+
                            ') ratio_populations must be a list of two Population objects')
            for i in range(2):
                if not isinstance(ratio_populations[i], Population):
                    raise TypeError('Adder(' + self.name +
                                    ') ratio_populations must be a list of two Population objects')
        self.ratio_populations = ratio_populations

    def update_expectation(self):
        """
        do addition for expectation
        """
        newcomers = 0
        if len(self.from_population.future) > 0:
            newcomers = self.from_population.future[0]

        if newcomers > 0:
            if getattr(self, "scale_factor", None) is not None:
                newcomers = self.scale_factor.get_value() * newcomers
            if getattr(self, "ratio_populations", None) is not None:
                if self.ratio_populations[1].history[-1] > 0.:
                    ratio = self.ratio_populations[0].history[-1]/self.ratio_populations[1].history[-1]
                    newcomers = ratio * newcomers

            self.to_population.update_future_fast(newcomers)

    def update_data(self):
        """
        do addition for data
        """
        newcomers = 0
        if len(self.from_population.future) > 0:
            newcomers = self.from_population.future[0]

        if newcomers > 0:
            scale = 1
            if getattr(self,"scale_factor",None) is not None:
                scale = self.scale_factor.get_value() * scale
            if getattr(self, "ratio_populations", None) is not None:
                if self.ratio_populations[1].history[-1] > 0.:
                    ratio = 1. * self.ratio_populations[0].history[-1] / self.ratio_populations[1].history[-1]
                    scale = ratio * scale

            if scale != 1:
                iscale = int(scale)
                i_newcomers = iscale*newcomers
                fscale = scale - iscale
                f_newcomers = stats.binom.rvs(newcomers, fscale)
                newcomers = i_newcomers + f_newcomers

            self.to_population.update_future_fast(newcomers)




# -*- coding: utf-8 -*-
"""
Trajectory: Given a model: characterize the relation between the model transmission rate parameter (alpha) and the
infection trajectory growth rate (delta: fractional growth per day).

This is evaluated by running the model for different values of the growth rate parameter.
The interpolation function is done as delta(ln_a), but that is hidden from user

After doing so, the following methods can be called:
- get_delta(alpha)
- get_alpha(delta)

It is assumed that the model has a consistent relation between alpha and delta throughout (time delays not changing)

@author: karlen
"""

import copy
from scipy.interpolate import interp1d
import numpy as np

class Trajectory:
    """ Trajectory: relationship between model transmission rate parameter and infection trajectory growth parameter
    """

    def __init__(self, model, contagious_pop_name: str, rate_modifier_name: str, rate_range: list):
        self.model = model
        self.contagious_pop_name = contagious_pop_name
        self.rate_modifier_name = rate_modifier_name
        if rate_range is None:
            raise TypeError('Error in constructing trajectory'+
                            ': rate_range cannot be None')
        if not isinstance(rate_range, list):
            raise TypeError('Error in constructing trajectory'+
                            ': rate_range must be a list')
        if len(rate_range) != 2:
            raise TypeError('Error in constructing trajectory'+
                            ': rate_range must be a list of length 2')
        if rate_range[0] <= 0. or rate_range[1] <= 0. or rate_range[1] <= rate_range[0]:
            raise ValueError('Error in constructing trajectory'+
                            ': rate_range list elements must be positive and second element larger than first')
        self.rate_range = rate_range
        self.log_alphas = None
        self.deltas = None
        self.delta_inter = None
        self.log_alpha_inter = None

        self.calc_deltas()

    def calc_deltas(self):
        # use 100 points for the interpolation (using log alpha)
        n_point = 100
        length = np.log(self.rate_range[1]) - np.log(self.rate_range[0])
        self.log_alphas = np.arange(np.log(self.rate_range[0]), np.log(self.rate_range[1]), length/n_point)

        # make a copy of model, since it needs to be modified
        tr_model = copy.deepcopy(self.model)
        rate_modifier = tr_model.transitions[self.rate_modifier_name]
        rate_step = rate_modifier.trigger_step
        # turn off all transitions on or after the rate_modifier
        for trans_name in tr_model.transitions:
            if trans_name != self.rate_modifier_name:
                transition = tr_model.transitions[trans_name]
                if transition.trigger_step >= rate_step:
                    transition.enabled = False

        rate_par = rate_modifier.parameter_after

        # run the simulation 20 steps after the transition - use the last points to work out growth rate
        n_step = rate_step + 20

        self.deltas = []
        for log_alpha in self.log_alphas:
            alpha = np.exp(log_alpha)
            rate_par.set_value(alpha)
            tr_model.reset()
            tr_model.evolve_expectations(n_step)
            ratio = tr_model.populations[self.contagious_pop_name].history[n_step] / tr_model.populations[self.contagious_pop_name].history[n_step-1]
            delta = ratio - 1
            self.deltas.append(delta)

        self.log_alpha_inter = interp1d(self.deltas, self.log_alphas, kind='cubic')
        self.delta_inter = interp1d(self.log_alphas, self.deltas, kind='cubic')

    def get_delta(self, alpha):
        return self.delta_inter(np.log(alpha)).item()

    def get_alpha(self, delta):
        return np.exp(self.log_alpha_inter(delta).item())


# -*- coding: utf-8 -*-
"""
Produce .csv file suitable for submission to covid19-forecast-hub (https://github.com/reichlab/covid19-forecast-hub)



@author: karlen
"""
from pathlib import Path
import numpy as np
import pandas as pd
from scipy import stats
import requests
import datetime
from datetime import timedelta
import copy

from pypmca import Model, Population, Delay, Parameter, Multiplier, Propagator, \
    Splitter, Adder, Subtractor, Chain, Modifier, Injector, Ensemble


class Forecast_hub:
    """ Forecast_hub: produce .csv file for covid19-forecast-hub
    """

    def __init__(self, folder, model_names):

        self.model_dir = Path(folder).resolve()
        self.model_names = model_names
        self.pd_dict = self.get_data()
        self.buff = [['forecast_date', 'target', 'target_end_date', 'location', 'type', 'quantile', 'value']]

        # number of daysfor case_normalization
        self.case_norm = 14
        # number of days for death_normalization
        self.death_norm = 14

        self.fips_code = {
            'AL': '01',
            'AK': '02',
            'AZ': '04',
            'AR': '05',
            'CA': '06',
            'CO': '08',
            'CT': '09',
            'DE': '10',
            'DC': '11',
            'FL': '12',
            'GA': '13',
            'HI': '15',
            'ID': '16',
            'IL': '17',
            'IN': '18',
            'IA': '19',
            'KS': '20',
            'KY': '21',
            'LA': '22',
            'ME': '23',
            'MD': '24',
            'MA': '25',
            'MI': '26',
            'MN': '27',
            'MS': '28',
            'MO': '29',
            'MT': '30',
            'NE': '31',
            'NV': '32',
            'NH': '33',
            'NJ': '34',
            'NM': '35',
            'NY': '36',
            'NC': '37',
            'ND': '38',
            'OH': '39',
            'OK': '40',
            'OR': '41',
            'PA': '42',
            'PR': '72',
            'RI': '44',
            'SC': '45',
            'SD': '46',
            'TN': '47',
            'TX': '48',
            'UT': '49',
            'VT': '50',
            'VA': '51',
            'WA': '53',
            'WV': '54',
            'WI': '55',
            'WY': '56'
        }

    def get_data(self):
        data_folder = 'data/covid19/USA'

        success = True
        pd_dict = {}
        try:
            data_desc_resp = requests.get('http://data.ipypm.ca/get_data_desc/' + data_folder)
        except requests.exceptions.RequestException as error:
            print('Error retrieving data description over network:')
            success = False
        if success:
            data_description = data_desc_resp.json()
            data_description['folder'] = data_folder
            print(data_description['description'])

            # load the data into a panda dictionary
            for filename in data_description['files']:
                path = data_folder + '/' + filename
                success = True
                try:
                    csv_resp = requests.get('http://data.ipypm.ca/get_csv/' + path, stream=True)
                except requests.exceptions.RequestException as error:
                    print('Error retrieving data over network:')
                    print()
                    print(error)
                    success = False

                if success:
                    pd_dict[filename] = pd.read_csv(csv_resp.raw)

        return pd_dict

    def test_Model_copy_values_from(self):
        path_model_2 = self.model_dir / 'ref_model_2.pypm'
        ref_model_2 = Model.open_file(path_model_2)

    def get_csv(self,forecast_date,us_deaths):
        t0 = datetime.date(2020, 3, 1)
        forecast_date_text = forecast_date.isoformat()
        day_of_week = forecast_date.weekday()
        days_after_t0 = (forecast_date - t0).days
        first_sunday = days_after_t0
        if day_of_week == 0:
            first_sunday -= 1
        elif day_of_week < 6:
            first_sunday += 6 - day_of_week

        names = ['case', 'death', 'hosp']
        dict_names = ['case', 'death', 'hospitalization']
        periods = ['wk', 'wk', 'day']
        inc_types_list = [['inc'], ['inc', 'cum'], ['inc']]

        # collect information for full US
        us_inc_point_est_dict = {}
        us_inc_periods_dict = {}
        us_cum_point_est_dict = {}
        us_cum_periods_dict = {}
        for dict_name in dict_names:
            us_inc_point_est_dict[dict_name] = {}
            us_inc_periods_dict[dict_name] = {}
        us_cum_point_est_dict['death'] = {}
        us_cum_periods_dict['death'] = {}
        state_deaths = 0

        for state in self.fips_code:
        #for state in ['TX','SC','FL']:
            abbrev = state.lower()
            location = self.fips_code[state]
            deaths = self.pd_dict['usa-jhu-pypm.csv'][abbrev.upper() + '-dt'].fillna(0).values

            success = False
            for model_name in self.model_names:
                try:
                    filename = abbrev + model_name + '.pypm'
                    path_model = self.model_dir / filename
                    case_model = Model.open_file(path_model)
                    success = True
                    break
                except:
                    pass
            for model_name in self.model_names:
                for model_suffix in ['_d','_h','']:
                    try:
                        filename = abbrev + model_name + model_suffix + '.pypm'
                        path_model = self.model_dir / filename
                        death_model = Model.open_file(path_model)
                        break
                    except:
                        pass
            for model_name in self.model_names:
                for model_suffix in ['_h','']:
                    try:
                        filename = abbrev + model_name + model_suffix + '.pypm'
                        path_model = self.model_dir / filename
                        hosp_model = Model.open_file(path_model)
                        break
                    except:
                        pass
            if not success:
                raise RuntimeError('No model for: ', abbrev)

            models = [case_model, death_model, hosp_model]
            print(models[0].name, models[1].name, models[2].name)

            for i in range(3):
                model = models[i]
                inc_types = inc_types_list[i]

                hub_dict = model.user_dict['forecast_hub'][forecast_date][dict_names[i]]
                point_est_dict = hub_dict['point_estimates']
                quantile_dict = hub_dict['quantiles']
                inc_periods = hub_dict['inc_periods']
                cum_periods = [deaths[first_sunday-1]] * len(inc_periods[0])
                for inc_type in inc_types:
                    # for cum record (so far, only deaths needs this)
                    sum = deaths[first_sunday-1]
                    if inc_type == 'cum':
                        state_deaths += sum
                    for i_period in point_est_dict:
                        index = int(i_period) - 1
                        target = i_period+' '+periods[i]+' ahead '+inc_type+' '+names[i]
                        days = int(i_period)
                        if periods[i] == 'wk':
                            days *= 7
                        else:
                            days += 1
                        target_end_date = (t0 + timedelta(days=(first_sunday + days - 1))).isoformat()

                        # point estimates
                        value = point_est_dict[i_period]
                        if inc_type == 'cum':
                            sum += point_est_dict[i_period]
                            value = sum

                            if i_period in us_cum_point_est_dict['death']:
                                us_cum_point_est_dict['death'][i_period] += sum
                            else:
                                us_cum_point_est_dict['death'][i_period] = sum

                        self.add_record(forecast_date_text,target,target_end_date,location,'point','NA',value)

                        # quantiles
                        if inc_type == 'inc':
                            if i_period in us_inc_point_est_dict[dict_names[i]]:
                                us_inc_point_est_dict[dict_names[i]][i_period] += value
                            else:
                                us_inc_point_est_dict[dict_names[i]][i_period] = value

                            for quant in quantile_dict[i_period]:
                                value = quantile_dict[i_period][quant]
                                self.add_record(forecast_date_text, target, target_end_date, location, 'quantile', quant, value)

                            if i_period in us_inc_periods_dict[dict_names[i]]:
                                ip_len = len(inc_periods[index])
                                for i_rep in range(len(us_inc_periods_dict[dict_names[i]][i_period])):
                                    # in case there are fewer repetitions, loop over others:
                                    ip_index = i_rep % ip_len
                                    us_inc_periods_dict[dict_names[i]][i_period][i_rep] += inc_periods[index][ip_index]
                            else:
                                us_inc_periods_dict[dict_names[i]][i_period] = copy.copy(inc_periods[index])
                        elif inc_type == 'cum':
                            for i_rep in range(len(inc_periods[index])):
                                cum_periods[i_rep] += inc_periods[index][i_rep]
                            for quant in quantile_dict[i_period]:
                                value = np.percentile(cum_periods, float(quant)*100.)
                                self.add_record(forecast_date_text, target, target_end_date, location, 'quantile', quant, value)

                            if i_period in us_cum_periods_dict[dict_names[i]]:
                                ip_len = len(cum_periods)
                                for i_rep in range(len(us_cum_periods_dict[dict_names[i]][i_period])):
                                    # in case there are fewer repetitions, loop over others:
                                    ip_index = i_rep % ip_len
                                    us_cum_periods_dict[dict_names[i]][i_period][i_rep] += cum_periods[ip_index]
                            else:
                                us_cum_periods_dict[dict_names[i]][i_period] = [cum_periods[i_rep] for i_rep in range(len(cum_periods))]

        # return 'US' summary:
        # there are additional regions used to define total US deaths Need to correct by adding the additional deaths here
        additional_deaths = us_deaths - state_deaths
        print('US total: additional deaths included:', additional_deaths)

        location = 'US'
        for i in range(3):
            inc_types = inc_types_list[i]
            if dict_names[i] in ['case']:
                quants = [0.025, 0.1, 0.25, 0.5, 0.75, 0.9, 0.975]
            elif dict_names[i] in ['death', 'hospitalization']:
                quants = [0.01, 0.025] + [0.05 + 0.05 * i for i in range(19)] + [0.975, 0.99]
            for inc_type in inc_types:
                for i_period in us_inc_point_est_dict[dict_names[i]]:
                    target = i_period + ' ' + periods[i] + ' ahead ' + inc_type + ' ' + names[i]
                    days = int(i_period)
                    if periods[i] == 'wk':
                        days *= 7
                    else:
                        days += 1
                    target_end_date = (t0 + timedelta(days=(first_sunday + days - 1))).isoformat()

                    # point estimates
                    value = us_inc_point_est_dict[dict_names[i]][i_period]
                    if inc_type == 'cum':
                        value = us_cum_point_est_dict[dict_names[i]][i_period] + additional_deaths
                    self.add_record(forecast_date_text, target, target_end_date, location, 'point', 'NA', value)

                    # quantiles
                    if inc_type == 'inc':
                        for quant in quants:
                            value = np.percentile(us_inc_periods_dict[dict_names[i]][i_period], float(quant)*100.)
                            quant_text = '{0:0.3f}'.format(quant)
                            self.add_record(forecast_date_text, target, target_end_date, location, 'quantile', quant_text,
                                            value)

                    elif inc_type == 'cum':
                        for quant in quants:
                            value = np.percentile(us_cum_periods_dict[dict_names[i]][i_period], float(quant) * 100.) + additional_deaths
                            quant_text = '{0:0.3f}'.format(quant)
                            self.add_record(forecast_date_text, target, target_end_date, location, 'quantile', quant_text,
                                            value)

        return self.buff



    def add_record(self, forecast_date, target, target_end_date, location, record_type, quantile_text, value):
        record = [forecast_date, target, target_end_date, location, record_type, quantile_text]
        record.append('{0:0.1f}'.format(value))
        self.buff.append(record)

my_forecast = Forecast_hub('/Users/karlen/pypm-temp/usa', ['_2_3_0906'])
# Indicate the total US deaths (up to and including Saturday) here:
us_deaths = 188538

my_csv = my_forecast.get_csv(datetime.date(2020, 9, 6), us_deaths)
pass
with open('/Users/karlen/pypm-temp/test-forecast.csv','w') as out:
    for line in my_csv:
        record = ','.join(line)
        out.write(record + '\n')



#!/usr/bin/env python

"""Tests for `pypmca` package."""

import pytest

from pypmca import Model, Population, Delay, Parameter, Multiplier, Propagator, \
    Splitter, Adder, Subtractor, Chain, Modifier, Injector, Ensemble
from pypmca.analysis.Optimizer import Optimizer
from pypmca.analysis.Trajectory import Trajectory
import numpy as np
import copy
from pathlib import Path

example_dir = Path('../../examples/').resolve()
path_model_2_5 = example_dir / 'ref_model_2_5.pypm'
path_model_2_4 = example_dir / 'ref_model_2_4.pypm'
path_model_2_3 = example_dir / 'ref_model_2_3.pypm'
path_model_2_2 = example_dir / 'ref_model_2_2.pypm'
path_model_2 = example_dir / 'ref_model_2.pypm'
path_model_1 = example_dir / 'ref_model_1.pypm'

def test_Model_copy_values_from():
    ref_model_2 = Model.open_file(path_model_2)
    alberta_model_2 = Model.open_file('ab_2_0514.pypm')

    ref_model_2.copy_values_from(alberta_model_2)

    ref_model_2.reset()
    ref_model_2.evolve_expectations(80)
    alberta_model_2.reset()
    alberta_model_2.evolve_expectations(80)

    for pop_name in ref_model_2.populations:
        pop = ref_model_2.populations[pop_name]
        # Bug fixed in ref_model_2, not in alberta_model_2 accounts for small differences in icu
        if 'vent' not in pop.name and 'in_' not in pop.name:
            ref_hist = pop.history
            alberta_hist = alberta_model_2.populations[pop_name].history
            for i in range(len(ref_hist)):
                if np.abs(ref_hist[i] - alberta_hist[i]) > 0.0001:
                    print(pop.name, i, ref_hist[i], alberta_hist[i], np.abs(ref_hist[i] - alberta_hist[i]))
                assert np.abs(ref_hist[i] - alberta_hist[i]) < 0.001

def test_Model_properties():
    """tests to ensure the properties of Model"""
    ref_model_2 = Model.open_file(path_model_2)
    ref_model_1 = Model.open_file(path_model_1)

    for test_model in [ref_model_1, ref_model_2]:
        # check simple scaling of initial contagious population
        # Will not be exact due to bootstrap
        EPS = 0.01
        cont_0 = test_model.parameters['cont_0'].get_value()
        test_model.reset()
        test_model.evolve_expectations(200)
        n_reported1 = test_model.populations['reported'].history[-1]
        assert test_model.populations['contagious'].history[-1] < 10.

        test_model.parameters['cont_0'].set_value(2 * cont_0)
        test_model.reset()
        test_model.evolve_expectations(200)
        n_reported2 = test_model.populations['reported'].history[-1]
        assert np.abs(n_reported2 - n_reported1 * 2) < EPS * n_reported1 * 2

    # check that the mean of many data runs is near expectation

    ref_model = Model.open_file(path_model_2)
    ref_model.reset()
    n_days = 60
    ref_model.evolve_expectations(n_days)
    sim_models = []
    n_rep = 100
    for i in range(n_rep):
        sim_model = copy.deepcopy(ref_model)
        sim_model.reset()
        sim_model.generate_data(n_days)
        sim_models.append(sim_model)
    for pop_name in ref_model.populations:
        pop = ref_model.populations[pop_name]
        if pop.show_sim:
            results = []
            for sim_model in sim_models:
                results.append(sim_model.populations[pop_name].history[-1])
            mean = np.mean(np.array(results))
            std = np.std(np.array(results))
            error = std / np.sqrt(1. * n_rep)
            expect = ref_model.populations[pop_name].history[-1]
            assert np.abs(expect - mean) < 8. * error


def test_Ensemble_properties_identical():
    """tests to ensure the properties of Ensemble"""
    # Test that the ensemble of two identical models
    # behalves like twice a single model.
    # independent of the contact matrix
    test_a = Model.open_file(path_model_2)
    test_a.name = 'test_a'
    test_b = Model.open_file(path_model_2)
    test_b.name = 'test_b'
    reference = Model.open_file(path_model_2)
    reference.name = 'reference'
    single = Model.open_file(path_model_2)
    single.name = 'single'

    test_ensemble = Ensemble('test_ensemble', reference)
    test_ensemble.upload_models([test_a, test_b])

    contacts = [
        [[1., 0.0], [0.0, 1.]],
        [[1., 1.0], [1.0, 1.]],
        [[1., 0.5], [0.7, 1.]],
    ]

    for contact in contacts:
        test_ensemble.define_cross_transmission('infection cycle', 'infected',
                                                'susceptible', 'total',
                                                'contagious', 'alpha',
                                                contact_type='fixed',
                                                contact=contact)

        n_days = 100
        test_ensemble.reset()
        test_ensemble.evolve_expectations(n_days)

        single.reset()
        single.evolve_expectations(n_days)

        for pop_name in test_ensemble.populations:
            pop = test_ensemble.populations[pop_name]
            if pop.show_sim:
                ens_hist = test_ensemble.populations[pop_name].history
                single_hist = single.populations[pop_name].history
                for i in range(len(ens_hist)):
                    ratio = ens_hist[i] / single_hist[i]
                    assert np.abs(ratio - 2.) < 0.001

def test_Ensemble_properties_equality():
    """tests to ensure the properties of Ensemble"""
    # Test that the ensemble of two identical models, with different scales
    # and equality contact matrix behaves like the same ensemble
    # with independent contact matrix
    test_a1 = Model.open_file(path_model_2_3)
    test_a1.name = 'test_a1'
    test_b1 = Model.open_file(path_model_2_3)
    test_b1.name = 'test_b1'
    reference1 = Model.open_file(path_model_2_3)
    reference1.name = 'reference1'
    cont_0 = test_b1.parameters['cont_0'].get_value()
    test_b1.parameters['cont_0'].set_value(cont_0*2.)
    N_0 = test_b1.parameters['N_0'].get_value()
    test_b1.parameters['N_0'].set_value(N_0*2)

    test_ensemble1 = Ensemble('test_ensemble1', reference1)
    test_ensemble1.upload_models([test_a1, test_b1])

    test_ensemble1.define_cross_transmission('infection cycle', 'infected',
                                            'susceptible', 'total',
                                            'contagious', 'alpha',
                                            contact_type='diagonal')

    test_a2 = Model.open_file(path_model_2_3)
    test_a2.name = 'test_a2'
    test_b2 = Model.open_file(path_model_2_3)
    test_b2.name = 'test_b2'
    reference2 = Model.open_file(path_model_2_3)
    reference2.name = 'reference2'
    cont_0 = test_b2.parameters['cont_0'].get_value()
    test_b2.parameters['cont_0'].set_value(cont_0*2.)
    N_0 = test_b2.parameters['N_0'].get_value()
    test_b2.parameters['N_0'].set_value(N_0*2)

    test_ensemble2 = Ensemble('test_ensemble2', reference2)
    test_ensemble2.upload_models([test_a2, test_b2])

    test_ensemble2.define_cross_transmission('infection cycle', 'infected',
                                            'susceptible', 'total',
                                            'contagious', 'alpha',
                                            contact_type='equality')

    n_days = 100

    test_ensemble1.reset()
    test_ensemble1.evolve_expectations(n_days)

    test_ensemble2.reset()
    test_ensemble2.evolve_expectations(n_days)

    for pop_name in test_ensemble1.populations:
        pop1 = test_ensemble1.populations[pop_name]
        if pop1.show_sim:
            pop1h = test_ensemble1.populations[pop_name].history
            pop2h = test_ensemble2.populations[pop_name].history
            for i in range(len(pop1h)):
                if pop2h[i] > 0.:
                    ratio = pop1h[i]/pop2h[i]
                    assert np.abs(ratio - 1.) < 0.03

def test_Ensemble_properties_different():
    """tests to ensure the properties of Ensemble with different sub models"""
    # Test that the ensemble of two identical models
    # behalves like twice a single model.
    # Only for the independent (ie. diagonal) case
    # Note: this test would fail if a fixed contact_matrix
    # is passed which happens to be the identity matrix
    # The difference is that when specified as independent
    # each model is booted independently
    # If a contact matrix is specified, then the boot goal
    # is the combined total of all models. So there will be a different starting point.
    test_a = Model.open_file(path_model_2)
    test_a.name = 'test_a'
    test_b = Model.open_file(path_model_2)
    test_b.name = 'test_b'
    test_b.parameters['alpha_0'].set_value(0.7)
    test_c = Model.open_file(path_model_2)
    test_c.name = 'test_c'
    test_d = Model.open_file(path_model_2)
    test_d.name = 'test_d'
    test_d.parameters['alpha_0'].set_value(0.7)
    reference = Model.open_file(path_model_2)
    reference.name = 'reference'

    test_ensemble = Ensemble('test_ensemble', reference)
    test_ensemble.upload_models([test_a, test_b])
    test_ensemble.define_cross_transmission('infection cycle', 'infected',
                                            'susceptible', 'total',
                                            'contagious', 'alpha',
                                            contact_type='diagonal')

    n_days = 100
    test_c.reset()
    test_c.evolve_expectations(n_days)
    test_d.reset()
    test_d.evolve_expectations(n_days)
    test_ensemble.reset()
    test_ensemble.evolve_expectations(n_days)
    for pop_name in test_ensemble.populations:
        pop = test_ensemble.populations[pop_name]
        if pop.show_sim:
            ens_hist = test_ensemble.populations[pop_name].history
            tc_hist = test_c.populations[pop_name].history
            td_hist = test_d.populations[pop_name].history
            for i in range(len(ens_hist)):
                ratio = ens_hist[i] / (tc_hist[i] + td_hist[i])
                assert np.abs(ratio - 1.) < 0.01

def test_point_estimate():
    start_day = 12
    end_day = 60
    ref_2 = Model.open_file(path_model_2_2)
    sim_2 = Model.open_file(path_model_2_2)

    # do fit of alpha_0, alpha_1, cont_0, trans_rate_1_time
    for par_name in ['alpha_0', 'alpha_1', 'cont_0']:
        par = ref_2.parameters[par_name]
        par.set_variable(None, None)

    par = ref_2.parameters['trans_rate_1_time']
    par.set_variable(None, None)
    par.set_min(13)
    par.set_max(19)

    sim_2.reset()
    sim_2.generate_data(end_day)
    optimizer = Optimizer(ref_2, 'total reported', sim_2.populations['reported'].history, [start_day, end_day])
    optimizer.reset_variables()

    scan_dict = optimizer.i_fit()
    assert ref_2.parameters['trans_rate_1_time'].get_value() in [15,16,17]

    par = ref_2.parameters['trans_rate_1_time']
    par.set_fixed()

    popt, pcov = optimizer.fit()
    assert np.abs(ref_2.parameters['alpha_0'].get_value()-ref_2.parameters['alpha_0'].initial_value) < 0.06
    assert np.abs(ref_2.parameters['alpha_1'].get_value() - ref_2.parameters['alpha_1'].initial_value) < 0.02
    assert np.abs(ref_2.parameters['cont_0'].get_value() - ref_2.parameters['cont_0'].initial_value) < 20.

def test_point_estimates_repeated():
    start_day = 12
    end_day = 60
    ref_2 = Model.open_file(path_model_2_2)
    sim_2 = Model.open_file(path_model_2_2)

    # do fit of alpha_0, alpha_1, cont_0
    par_names = ['alpha_0', 'alpha_1', 'cont_0']
    sums = {}
    sum2s = {}
    for par_name in par_names:
        par = ref_2.parameters[par_name]
        par.set_variable(None, None)
        sums[par_name] = 0.
        sum2s[par_name] = 0.

    n_rep = 10
    fit_stat_list = []
    for i in range(n_rep):
        sim_2.reset()
        sim_2.generate_data(end_day)
        optimizer = Optimizer(ref_2, 'total reported', sim_2.populations['reported'].history, [start_day, end_day])
        optimizer.reset_variables()
        popt, pcov = optimizer.fit()
        fit_stat_list.append(optimizer.fit_statistics)
        for par_name in par_names:
            value = ref_2.parameters[par_name].get_value()
            sums[par_name] += value
            sum2s[par_name] += value**2

    ass_std = {}
    ass_std['alpha_0'] = 0.03
    ass_std['alpha_1'] = 0.01
    ass_std['cont_0'] = 10.

    means = {}
    std = {}
    for par_name in par_names:
        means[par_name] = sums[par_name]/n_rep
        std[par_name] = np.sqrt(sum2s[par_name]/n_rep - means[par_name]**2)
        assert std[par_name] < ass_std[par_name]
        truth = ref_2.parameters[par_name].initial_value
        assert np.abs((means[par_name]-truth)/std[par_name]/np.sqrt(1.*n_rep)) < 3.

    ndof = fit_stat_list[0]['ndof']
    chi2_list = [fit_stat_list[i]['chi2'] for i in range(n_rep)]
    chi2_mean = np.mean(chi2_list)
    assert np.abs(chi2_mean - ndof) < 8.
    acor_list = [fit_stat_list[i]['acor'] for i in range(n_rep)]
    acor_mean = np.mean(acor_list)
    assert np.abs(acor_mean) < 0.2

def test_sim_gof():
    start_day = 12
    end_day = 60
    ref_2 = Model.open_file(path_model_2_2)
    sim_2 = Model.open_file(path_model_2_2)

    # do fit of alpha_0, alpha_1, cont_0
    par_names = ['alpha_0', 'alpha_1', 'cont_0']
    for par_name in par_names:
        par = ref_2.parameters[par_name]
        par.set_variable(None, None)

    sim_2.reset()
    sim_2.generate_data(end_day)
    optimizer = Optimizer(ref_2, 'total reported', sim_2.populations['reported'].history, [start_day, end_day])
    optimizer.reset_variables()
    popt, pcov = optimizer.fit()
    fit_statistics = optimizer.fit_statistics

    optimizer.calc_chi2s = False
    optimizer.calc_chi2f = True
    n_rep = 10
    optimizer.calc_sim_gof(n_rep)

    fit_stat_list = optimizer.fit_stat_list
    ndof = fit_stat_list[0]['ndof']
    chi2_list = [fit_stat_list[i]['chi2'] for i in range(n_rep)]
    chi2_mean = np.mean(chi2_list)
    assert np.abs(chi2_mean - ndof) < 8.
    acor_list = [fit_stat_list[i]['acor'] for i in range(n_rep)]
    acor_mean = np.mean(acor_list)
    assert np.abs(acor_mean) < 0.2

def test_report_noise():
    start_day = 12
    end_day = 80
    ref_2 = Model.open_file(path_model_2_2)
    ref_2.parameters['report_noise'].set_value(0.1)
    sim_2 = copy.deepcopy(ref_2)

    # do fit of alpha_0, alpha_1, cont_0
    par_names = ['alpha_0', 'alpha_1', 'cont_0']
    sums = {}
    sum2s = {}
    for par_name in par_names:
        par = ref_2.parameters[par_name]
        par.set_variable(None, None)
        sums[par_name] = 0.
        sum2s[par_name] = 0.

    n_rep = 10
    fit_stat_list = []
    for i in range(n_rep):
        sim_2.reset()
        sim_2.generate_data(end_day)
        optimizer = Optimizer(ref_2, 'total reported', sim_2.populations['reported'].history, [start_day, end_day])
        optimizer.reset_variables()
        popt, pcov = optimizer.fit()
        fit_stat_list.append(optimizer.fit_statistics)
        for par_name in par_names:
            value = ref_2.parameters[par_name].get_value()
            sums[par_name] += value
            sum2s[par_name] += value**2

    ass_std = {}
    ass_std['alpha_0'] = 0.03
    ass_std['alpha_1'] = 0.01
    ass_std['cont_0'] = 10.

    means = {}
    std = {}
    for par_name in par_names:
        means[par_name] = sums[par_name]/n_rep
        std[par_name] = np.sqrt(sum2s[par_name]/n_rep - means[par_name]**2)
    for par_name in par_names:
        assert std[par_name] < ass_std[par_name]
        truth = ref_2.parameters[par_name].initial_value
        assert np.abs((means[par_name]-truth)/std[par_name]/np.sqrt(1.*n_rep)) < 3.

    ndof = fit_stat_list[0]['ndof']
    chi2_list = [fit_stat_list[i]['chi2'] for i in range(n_rep)]
    chi2_mean = np.mean(chi2_list)
    assert chi2_mean - ndof > 100.
    acor_list = [fit_stat_list[i]['acor'] for i in range(n_rep)]
    acor_mean = np.mean(acor_list)
    assert acor_mean < -0.3

def test_report_noise_days():
    start_day = 12
    end_day = 80
    ref_2 = Model.open_file(path_model_2_2)
    ref_2.parameters['report_noise'].set_value(0.1)
    # BC: no reporting on Sundays
    ref_2.parameters['report_days'].set_value(63)
    sim_2 = copy.deepcopy(ref_2)

    # do fit of alpha_0, alpha_1, cont_0
    par_names = ['alpha_0', 'alpha_1', 'cont_0']
    sums = {}
    sum2s = {}
    for par_name in par_names:
        par = ref_2.parameters[par_name]
        par.set_variable(None, None)
        sums[par_name] = 0.
        sum2s[par_name] = 0.

    n_rep = 10
    fit_stat_list = []
    for i in range(n_rep):
        sim_2.reset()
        sim_2.generate_data(end_day)
        optimizer = Optimizer(ref_2, 'total reported', sim_2.populations['reported'].history, [start_day, end_day])
        optimizer.reset_variables()
        popt, pcov = optimizer.fit()
        fit_stat_list.append(optimizer.fit_statistics)
        for par_name in par_names:
            value = ref_2.parameters[par_name].get_value()
            sums[par_name] += value
            sum2s[par_name] += value ** 2

    ass_std = {}
    ass_std['alpha_0'] = 0.05
    ass_std['alpha_1'] = 0.01
    ass_std['cont_0'] = 10.

    means = {}
    std = {}
    for par_name in par_names:
        means[par_name] = sums[par_name] / n_rep
        std[par_name] = np.sqrt(sum2s[par_name] / n_rep - means[par_name] ** 2)
    for par_name in par_names:
        assert std[par_name] < ass_std[par_name]
        truth = ref_2.parameters[par_name].initial_value
        assert np.abs((means[par_name] - truth) / std[par_name] / np.sqrt(1. * n_rep)) < 3.

    ndof = fit_stat_list[0]['ndof']
    chi2_list = [fit_stat_list[i]['chi2'] for i in range(n_rep)]
    chi2_mean = np.mean(chi2_list)
    assert chi2_mean/ndof > 2.5
    acor_list = [fit_stat_list[i]['acor'] for i in range(n_rep)]
    acor_mean = np.mean(acor_list)
    assert acor_mean < -0.2

def test_trajectory():
    ref_2 = Model.open_file(path_model_2_2)
    trajectory = Trajectory(ref_2, 'contagious', 'trans_rate_1', [0.03, 0.75])
    alpha_c = trajectory.get_alpha(0.)
    delta_1 = trajectory.get_delta(0.1)
    delta_2 = trajectory.get_delta(0.2)
    assert np.abs(alpha_c - 0.152) < 0.001
    assert np.abs(trajectory.get_delta(alpha_c)) < 0.00001
    assert np.abs(delta_1 + 0.0513) < 0.0001
    assert np.abs(delta_2-0.0389) < 0.0001

def test_linear_modifier():
    ref_2_4 = Model.open_file(path_model_2_4)
    trajectory = Trajectory(ref_2_4, 'contagious', 'trans_rate_1', [0.02, 2.0])
    alpha_c = trajectory.get_alpha(0.)
    ref_2_4.parameters['alpha_1'].set_value(alpha_c)
    ref_2_4.parameters['to_icu_delay_mean'].set_value(0.5)
    ref_2_4.parameters['to_icu_delay_sigma'].set_value(0.5)
    ref_2_4.transitions['mod_icu_frac'].enabled = True
    ref_2_4.parameters['icu_frac_time'].set_value(60)
    ref_2_4.parameters['icu_frac_0'].set_value(0.1)
    ref_2_4.parameters['icu_frac_slope'].set_value(0.01)
    ref_2_4.parameters['icu_frac_nstep'].set_value(10)
    ref_2_4.reset()
    ref_2_4.evolve_expectations(100)
    pop_hist = ref_2_4.populations['icu admissions'].history
    assert np.abs(pop_hist[55]-pop_hist[54]-pop_hist[50]+pop_hist[49]) < 0.01
    assert np.abs(pop_hist[75] - pop_hist[74] - pop_hist[70] + pop_hist[69]) < 0.1
    assert np.abs((pop_hist[75] - pop_hist[74])/(pop_hist[55]-pop_hist[54]) - 2.) < 0.1

def test_model_2_5():
    ref_2_5 = Model.open_file(path_model_2_5)
    ref_2_3 = Model.open_file(path_model_2_3)
    #ref_2_5.transitions['vaccination'].enabled = True
    #ref_2_5.parameters['vaccination_number'].set_value(10000.)
    ref_2_5.reset()
    ref_2_5.evolve_expectations(200)
    ref_2_3.reset()
    ref_2_3.evolve_expectations(200)

    for pop_name in ref_2_3.populations:
        pop = ref_2_3.populations[pop_name]
        if pop.show_sim:
            print(pop_name)
            hist_2_3 = ref_2_3.populations[pop_name].history
            hist_2_5 = ref_2_5.populations[pop_name].history
            for i in range(len(hist_2_3)):
                if hist_2_3[i] > 0:
                    ratio = hist_2_5[i] / hist_2_3[i]
                    assert np.abs(ratio - 1.) < 0.001

def test_max_mult():
    ref_2_5 = Model.open_file(path_model_2_5)
    ref_2_5.transitions['vaccination'].enabled = True
    ref_2_5.parameters['vaccination_number'].set_value(800000.)
    ref_2_5.parameters['vaccination_time'].set_value(5)
    ref_2_5.reset()
    ref_2_5.boot(expectations=True)
    ref_2_5.evolve_expectations(200)
    i=1


# -*- coding: utf-8 -*-
"""
Population: A class that holds the state of a population as it evolves in time
as well as future contributions to the population:
    - history: a list containing the population state at each time step.
      The first element of the list corresponds to time = t0.
    - future: a list containing future additions to the population.
      The first element corresponds to the last time step of the history

A population can be used for tracking expectation values (reals) or
simulated data (integers).

@author: karlen
"""

from scipy import stats
from datetime import timedelta

from pypmca.Parameter import Parameter


class Population:
    """A class that keeps state information of a population at each time step:
        - population_name : string, short descriptor. Must be unique
        - initial_value : integer, float, or Parameter object
        - description: string, optional - for documentation
        - hidden: don't show in lists of populations when keeping list short
        - color: color to plot
        - show_sim: is it appropriate to show simulated data for this distribution?
        - report_noise: should simulated data include additional noise due to reporting?
        - if True, then report_noise_par is the lower limit of range [low,1] for which a uniform
          random number is drawn to select the number from today that are reported today
          the remaining will be in future report (along with some fraction of tomorrows report)
        - if True, then report_backlog_par is the lower limit of range [low,1] for which a uniform
          random number is drawn to select the number from the backlog that are reported today
          the remaining will be in a future report
        - if True, and the integer parameter report_days is provided, use that to determine which
          days of week that reports are produced. In BC, reports are zero on Sunday and included in the
          next reporting day. Bit encoded, with Monday = bit 0... Sunday = bit 6. BC: report_days = 63
    """

    def __init__(self, population_name: str, initial_value, description: str = '',
                 hidden: bool = True, color: str = 'black', show_sim: bool = False, report_noise: bool = False,
                 report_noise_par: Parameter = None, report_backlog_par: Parameter = None, report_days: Parameter = None):
        """
        Constructor

        Parameters
        ----------
        population_name : string, short descriptor
        initial_value : integer, float, or Parameter object

        Returns
        -------
        None.

        """
        if population_name.find(',') > -1:
            raise ValueError('Error in constructing ' + self.name +
                             ': name cannot contain a comma.')
        self.name = str(population_name)
        self.description = str(description)
        self.history = None
        self.model = None
        self.parameters = {}
        self.__initialization_by_parameter = False
        if not isinstance(initial_value, (float, int)) and not isinstance(initial_value, Parameter):
            raise TypeError('Error setting initial_value to population (' +
                            self.name + ') - must be a float, int, or Parameter object')
        self.initial_value = None
        self.set_initial_value(initial_value)

        self.future = []
        self.color = color
        self.hidden = hidden
        # identify those populations for which daily contributions are meaningful
        # this is set to False if a subtraction is performed on the population (see Subtractor.py)
        self.monotonic = True
        # identify those populations for which simulated data is appropriate to show 
        self.show_sim = show_sim
        # for tracking missed reports from yesterday (reporting noise)
        self.missed_yesterday = 0

        self.__report_noise = None
        self.__report_noise_par = None
        self.__report_backlog_par = None
        self.__report_days = None
        self.set_report_noise(report_noise, report_noise_par, report_backlog_par, report_days)

    def set_model(self, model):
        # when a connector is added to a model, the populations are informed of the parent model
        # in order to get t0 and step size information
        self.model = model

    def set_report_noise(self, report_noise, report_noise_par, report_backlog_par, report_days):

        self.__report_noise = report_noise
        if report_noise and (report_noise_par is None or
                             not isinstance(report_noise_par, Parameter)):
            raise TypeError('Error setting report_noise_par in population (' +
                            self.name + ') - it must be a Parameter object')
        self.__report_noise_par = report_noise_par
        if report_noise and report_backlog_par is not None:
            if not isinstance(report_backlog_par, Parameter):
                raise TypeError('Error setting report_backlog_par in population (' +
                                self.name + ') - it must be a Parameter object')
            self.__report_backlog_par = report_backlog_par
            self.parameters['backlog_par'] = self.__report_backlog_par
        if report_noise and report_days is not None:
            if not isinstance(report_days, Parameter):
                raise TypeError('Error setting report_days in population (' +
                                self.name + ') - it must be a Parameter object')
            elif report_days.parameter_type != 'int':
                raise TypeError('Error setting report_days in population (' +
                                self.name + ') - it must be an integer Parameter object')
            else:
                self.__report_days = report_days
                self.parameters['report_days'] = self.__report_days

        if report_noise:
            self.parameters['noise_par'] = self.__report_noise_par
        else:
            if 'noise_par' in self.parameters:
                self.parameters.pop('noise_par')
            if 'backlog_par' in self.parameters:
                self.parameters.pop('backlog_par')
            if 'report_days' in self.parameters:
                self.parameters.pop('report_days')
        # in case parameter changed, update the model list of parameters
        # unfortunately population does not have link to model
        # self.model.update_lists()

    def get_report_noise(self):
        return {'report_noise':self.__report_noise,
                'report_noise_par':self.__report_noise_par,
                'report_backlog_par':self.__report_backlog_par,
                'report_days':self.__report_days}

    def set_initial_value(self, initial_value):
        if isinstance(initial_value, (float, int)):
            self.history = [initial_value]
            self.__initialization_by_parameter = False
            if 'init_value' in self.parameters:
                self.parameters.pop('init_value')
        elif isinstance(initial_value, Parameter):
            self.history = [initial_value.get_value()]
            self.__initialization_by_parameter = True
            self.parameters['init_value'] = initial_value
        self.initial_value = initial_value

    def __str__(self):
        return self.name

    def do_time_step(self, expectations=True):
        """Perform one step in time by incrementing population number from future.
        After doing so, remove the future element
        
        if we are reporting data (not expectations) and report_noise is true
        then add the missed reports from yesterday and miss some from today

        if report_days is not None, then save all reports on non-reporting days for
        the next day
        """
        next_value = 0
        if self.future is not None:

            if expectations or not self.__report_noise:
                if len(self.future) > 0:
                    next_value = self.future[0]
            else:
                incoming = 0
                if len(self.future) > 0:
                    incoming = self.future[0]
                # are we reporting today? (new member added in ref model 2_2)
                try:
                    reporting = self.__report_days is None
                except:
                    reporting = True
                if not reporting:
                    # check if today is a reporting day
                    t0 = self.model.t0
                    days_after = timedelta(days=int((len(self.history)+0.5)*self.model.get_time_step()))
                    today = t0 + days_after
                    day_of_week = today.weekday()
                    reporting = (self.__report_days.get_value() & 2**day_of_week) > 0
                if reporting:
                    # how many will be reported from today?
                    low_edge = self.__report_noise_par.get_value()
                    frac_report = stats.uniform.rvs(loc=low_edge, scale=1. - low_edge)
                    n_report = stats.binom.rvs(incoming, frac_report)

                    # how many will be reported from the backlog?
                    try:
                        low_edge = self.__report_backlog_par.get_value()
                        frac_backlog = stats.uniform.rvs(loc=low_edge, scale=1. - low_edge)
                        n_backlog = stats.binom.rvs(self.missed_yesterday, frac_backlog)
                    except:
                        n_backlog = self.missed_yesterday

                    next_value = n_report + n_backlog

                    self.missed_yesterday -= n_backlog
                    self.missed_yesterday += incoming - n_report
                else:
                    self.missed_yesterday += incoming
                    next_value = 0

        self.history.append(self.history[-1] + next_value)
        # don't let the population go negative
        if self.history[-1] < 0:
            self.history[-1] = 0
        # remove the future element
        if self.future is not None:
            if len(self.future) > 0:
                del self.future[0]

    def reset(self):
        """Remove history and future, reinitialize history
        """
        self.future = []
        if isinstance(self.initial_value, Parameter):
            self.history = [self.initial_value.get_value()]
        else:
            self.history = [self.initial_value]

        self.missed_yesterday = 0

    def remove_history(self):
        """Replace history with array of length 1
        """
        if len(self.history) > 0:
            current_value = self.history[-1]
            self.history = [current_value]

    def scale_history(self, scale, expectations=True):
        if len(self.history) > 0:
            for i in range(len(self.history)):
                if expectations:
                    self.history[i] *= scale
                else:
                    nu = self.history[i] * scale
                    self.history[i] = int(round(nu))

    def scale_future(self, scale, expectations=True):
        if len(self.future) > 0:
            for i in range(len(self.future)):
                if expectations:
                    self.future[i] *= scale
                else:
                    nu = self.future[i] * scale
                    self.future[i] = int(round(nu))

    def update_future_expectation(self, scale, delay):
        """Include future expections, growing the future array if needed
        """
        for i in range(len(delay.future_expectations)):
            dfe = delay.future_expectations[i] * scale
            if len(self.future) > i:
                self.future[i] += dfe
            else:
                self.future.append(dfe)

    def update_future_data(self, scale, delay):
        """Include future data, growing the future array if needed
        """
        dist = stats.multinomial.rvs(int(scale), delay.future_expectations)
        i = -1
        for n_future in dist:
            i += 1
            if len(self.future) > i:
                self.future[i] += n_future
            else:
                self.future.append(n_future)

    def update_future_fast(self, value):
        """Modify the immediate future
        """
        if len(self.future) > 0:
            self.future[0] += value
        else:
            self.future.append(value)
